/**************************************************************************
 *   MudClient - mudcl.c                                                  *
 *   Copyright (C) 1997  Christophe Roux                                  *
 *   This program is free software; you can redistribute it and/or modify *
 *   it under the terms of the GNU General Public License as published by *
 *   the Free Software Foundation; either version 1, or (at your option)  *
 *   any later version.                                                   *
 *                                                                        *
 *   This program is distributed in the hope that it will be useful,      *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 *   GNU General Public License for more details.                         *
 *                                                                        *
 *   You should have received a copy of the GNU General Public License    *
 *   along with this program; if not, write to the Free Software          *
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
 *                                                                        *
 *   Apr 1997      bugs report : <jacker@ucfma.univ-bpclermont.fr>        *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <signal.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <curses.h>
#include <varargs.h>
#include <termios.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include "alias.h"
#include "cmd.h"
#include "mudcl.h"

#define VERSION  "1.2"
#undef DEBUG

WINDOW *screen,*status,*input;
int lin,col,cx,cy;
char *clear_line;

char line[HISTORY][LINE];
int pos[HISTORY];
int hist;

char current_line[LINE],vis_line[LINE];
int current_pos,vis_pos;
int current_hist;

int passmode;

char host[HOSTNAME+1];
int port;
char conn;

char path[FILENAME];

/**********************************************************************/
/*                            Socket function                         */
/**********************************************************************/

int call_socket (char *hostname, unsigned short portnum)
{
  struct sockaddr_in sa;
  struct hostent *hp;
  int s;

  if ((hp = gethostbyname (hostname)) == NULL) {
    return (-1);
  }
  memset (&sa, 0, sizeof (sa));
  memcpy ((char *) &sa.sin_addr, hp->h_addr, hp->h_length);
  sa.sin_family = hp->h_addrtype;
  sa.sin_port = htons ((u_short) portnum);
  if ((s = socket (hp->h_addrtype, SOCK_STREAM, 0)) < 0)
    return (-1);
  if (connect (s, (struct sockaddr *) &sa, sizeof sa) < 0) {
    close (s);
    return (-1);
  }
  return (s);
}

void send_crnl(int sock)
{
  char nl,cr;

  cr = '\r';
  nl = '\n';
  if ((int) write(sock,&cr,1) < 0)
    conn=0;
  if ((int) write(sock,&nl,1) < 0)
    conn=0;
}

/**********************************************************************/
/*                           ncurses functions                        */
/**********************************************************************/

void init()
{
  struct winsize window;
  int tty_des;
  int i;

  if ((tty_des = open("/dev/tty", O_RDWR, 0)) == -1) {
    fprintf(stderr,"Error : no tty\n");
    exit(0);
  }
  if (ioctl(tty_des, TIOCGWINSZ, &window) <0) {
    fprintf(stderr,"Error : ioctl\n");
    exit(0);
  }
  close(tty_des);
  lin = window.ws_row;
  col = window.ws_col;
  clear_line = (char *) malloc(col);
  for (i = 0 ; i < col ; i++)
    clear_line[i] = ' ';
  clear_line[col - 1] = '\0';
  cx = 0;
  cy = 0;
  passmode = 0;
  hist = 0;
  for (i = 0 ; i < HISTORY ; i++) {
    pos[i] = 0;
    line[i][0] = '\0';
  }
  current_line[0] = '\0';
  current_pos = 0;
  current_hist = 0;
  initscr();
  savetty();
  nonl();
  cbreak();
  noecho();
  screen = newwin(lin - 2,col,0,0);
  status = newwin(1,col,lin - 2,0);
  input  = newwin(1,col,lin - 1,0);
  scrollok(screen,TRUE);
  nodelay(input,TRUE);
}

void nospe(char *buf)
{
  while(*buf) {
    if (((unsigned char) *buf < (unsigned char) '!') && (*buf != '\n')&& (*buf != '\t'))
      *buf = ' ';
    if ((unsigned char)*buf == 0xFF) {
     *buf = ' ';
     buf++;
      if ((unsigned char)*buf == 0xFB) {
	passmode = 1;
	*buf = ' ';
      }
      else if ((unsigned char)*buf == 0xFC) {
	passmode = 0;
	*buf = ' ';
      }
    }
    buf++;
  }
}

void putscr(va_alist)
va_dcl
{
  va_list va;
  char *format;
  char buf[SIZE];

  va_start(va);
  format = va_arg(va,char *);
  vsprintf(buf,format,va);
  nospe(buf);
  mvwaddstr(screen,cy,cx,buf);
  getyx(screen,cy,cx);
  wrefresh(screen);
  va_end(va);
}

/**********************************************************************/
/*                            Screen function                         */
/**********************************************************************/

void scr(int sock)
{
  char buffer[SIZE],copy[SIZE * 2];
  char *src,*dst;
  int s;

  if ((s = read(sock,buffer,SIZE - 1)) <= 0)
    conn = 0;
  else {

    buffer[s] = '\0';
    src = buffer;
    dst = copy;
    while (*src)
    {
      *(dst++) = *(src++);
      if (*(dst - 1) == '%')
	*(dst++) = '%';
    }
    *dst = '\0';
    putscr(copy);
  }
}

/**********************************************************************/
/*                         Status bar function                        */
/**********************************************************************/

void statbar(va_alist)
va_dcl
{
  va_list va;
  char *format;
  char buf[SIZE];
  int i,l;

  va_start(va);
  format = va_arg(va,char *);
  vsprintf(buf,format,va);
  buf[col] = '\0';
  l = strlen(buf);
  for (i = l;i < col ;i++)
    buf[i] = ' ';
  wattron(status,A_REVERSE);
  mvwaddstr(status,0,0,buf);
  wattroff(status,A_REVERSE);
  wrefresh(status);
  va_end(va);
}

/**********************************************************************/
/*                           Keybord function                         */
/**********************************************************************/

void kbd(int sock)
{
  char c;
  int ptr_pos;
  int event_pos;
  int event;

  event = 1;
  c = mvwgetch(input,0,current_pos);
  switch(c) {
  case ERR:
  case 9:
    break;
  case 27:
    c = getch();
    if (c == 91) {
      switch(getch()) {
      case 0:
      case 144:
	break;
      case 65:
	if (hist == current_hist) {
	  strcpy(line[hist],current_line);
	  pos[hist] = current_pos;
	}
	hist--;
	if (hist < 0)
	  hist = HISTORY - 1;
	strcpy(current_line,line[hist]);
	current_pos = pos[hist];
	wclear(input);
	mvwaddstr(input,0,0,current_line);
	break;
      case 66:
	if (hist == current_hist) {
	  strcpy(line[hist],current_line);
	  pos[hist] = current_pos;
	}
	hist++;
	if (hist == HISTORY)
	  hist = 0;
	strcpy(current_line,line[hist]);
	current_pos = pos[hist];
	wclear(input);
	mvwaddstr(input,0,0,current_line);
	break;
      case 67:
	if (current_line[current_pos])
	  current_pos++;
	break;
      case 68:
	if (current_pos)
	  current_pos--;
	break;
      default:
	while (getch() != 126);
      }
    }
    break;
  case 0x0D:
    if (*current_line == HISTORY_CHAR)
    {
      for (event_pos = (current_hist - 1 + HISTORY) % HISTORY; 
	   (strncmp(current_line + 1,line[event_pos],strlen(current_line) - 1) 
	   && (event_pos != current_hist));
	   event_pos  = ((event_pos - 1) + HISTORY) % HISTORY);
      if (event_pos == current_hist)
      {
	putscr("\n*** Event not found\n");
	event = 0;
      }
      else
	strcpy(current_line,line[event_pos]);
    }
    if (event)
    {
      strcpy(line[current_hist],current_line);
      current_pos = strlen(current_line);
      pos[current_hist] = current_pos;
      current_hist++;
      hist = current_hist;
      if (current_hist == HISTORY)
	current_hist = 0;
      if (!is_cmd(current_line,sock))
	if (!is_alias(current_line,sock))
	{ 
	  if ((int) write(sock,current_line,current_pos) < 0)
	    conn=0;
	  send_crnl(sock);
	}
    }
    current_line[0] = '\0';
    current_pos = 0;
    wclear(input);
    break;
  case 8:
    if (current_pos) {
      current_pos--;
      for (ptr_pos = current_pos ; current_line[ptr_pos] ; ptr_pos++)
	current_line[ptr_pos] = current_line[ptr_pos + 1];
      current_line[ptr_pos] = current_line[ptr_pos + 1];
      if (!passmode)
      {
	mvwaddstr(input,0,0,clear_line);
	mvwaddstr(input,0,0,current_line);
      }
    }
    break;
  case 127:
    if (current_line[current_pos])
    {
      for (ptr_pos = current_pos ; current_line[ptr_pos] ; ptr_pos++)
	current_line[ptr_pos] = current_line[ptr_pos + 1];
      current_line[ptr_pos] = current_line[ptr_pos + 1];
      if (!passmode)
      {
	mvwaddstr(input,0,0,clear_line);
	mvwaddstr(input,0,0,current_line);
      }
    }
    break;
  default:
    if (current_line[current_pos])
    {
      for (ptr_pos = strlen(current_line) + 1 ; ptr_pos != current_pos ; ptr_pos--)
      {
	current_line[ptr_pos + 1] = current_line[ptr_pos];
      }
      current_line[ptr_pos + 1] = current_line[ptr_pos];
      current_line[current_pos] = c;
      current_pos++;      
    }
    else
    {
      current_line[current_pos] = c;
      current_pos++;
      current_line[current_pos] = '\0';
    }
    if (!passmode)
    {
      mvwaddstr(input,0,0,clear_line);
      mvwaddstr(input,0,0,current_line);
    }
    break;
  }
}

/**********************************************************************/
/*                            Client function                         */
/**********************************************************************/

void clt(int sock)
{
  fd_set rd,wd;
  struct timeval timeout;
  time_t sec;
  struct tm *loc_time;

  signal(SIGPIPE,SIG_IGN);
  FD_ZERO(&wd);
  FD_SET(sock,&wd);
  FD_SET(0,&wd);
  while(conn) {
    timeout.tv_sec = DELAY;
    timeout.tv_usec = UDELAY;
    rd=wd;
    select(NFDBITS,&rd,NULL,NULL,&timeout);
    if (FD_ISSET(sock,&rd))
      scr(sock);
    time(&sec);
    loc_time = localtime (&sec);
    statbar(" [%02d:%02d] %s %d ", loc_time->tm_hour, loc_time->tm_min, host, port);
    if (FD_ISSET(0,&rd))
      kbd(sock);
    wmove(input,0,current_pos);
    wrefresh(input);
  }
}

/**********************************************************************/
/*                             Main function                          */
/**********************************************************************/

void main(int argc,char **argv)
{
  int sock;

  if (getenv("MUDCL_PATH"))
    strcpy(path,getenv("MUDCL_PATH"));
  else
    strcpy(path,".");
  if (argc!=3) {
#ifdef DEBUG
    strcpy(host,"localhost");
    port = 4000;
#else
    fprintf(stderr,"Usage: %s hostname port\n",argv[0]);
    return;
#endif
  } else {
    strcpy(host,argv[1]);
    port = atoi(argv[2]);
  }
  if ((sock = call_socket(host,port))<0) {
    fprintf(stderr,"No connection at %s on %d\n",host,port);
    return;
  } else {
    init();
    if (!mudclrc(sock))
      putscr("*** Warning can't read ~/.mudclrc\n");
    putscr("*** Mud Client version %s\n",VERSION);
    putscr("*** Use -help to have help\n");
    putscr("*** Bugs reports Christophe Roux <jacker@ucfma.univ-bpclermont.fr>\n");
    putscr("*** Connected to %s\n",host);
    conn = 1;
    clt(sock);
  }
  resetty();
  endwin();
  fprintf(stderr,"Connection lost\n");
}
